from __future__ import annotations

from datetime import datetime
from typing import Any, Callable

import xarray
from xarray import DataArray, Dataset
from xclim.core import calendar

from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.models.cf_calendar import CfCalendar
from icclim.models.constants import PR, TAS, TAS_MAX, TAS_MIN
from icclim.models.frequency import Frequency, SliceMode
from icclim.models.netcdf_version import NetcdfVersion
from icclim.models.quantile_interpolation import QuantileInterpolation


class CfVariable:
    """
    CfVariable groups together two xarray DataArray for the same variable.
    One represent the whole studied period. The other is only the in base period used by
    percentile based indices to compute percentiles.

    Parameters
    ----------
    study_da: DataArray
        The variable studied.
    reference_da: DataArray
        The variable studied limited to the in base period.
    """

    name: str
    study_da: DataArray
    reference_da: DataArray

    def __init__(self, name: str, study_da: DataArray, reference_da: DataArray) -> None:
        self.name = name
        self.study_da = study_da
        self.reference_da = reference_da


class IndexConfig:
    """
    Configuration class for ECA&D indices.

    Parameters
    ----------
    freq: Frequency
        The expected resampling frequency of the output.
    cf_variables:
        List of CfVariable necessary to compute the index.
    save_percentile: bool = False
        On percentile based indices, if True, this saves the percentile in the output
        netcdf.
    is_percent:
        On indices resulting in a numbers of days, if True, this converts the results to
        % of the sampling frequency
    netcdf_version:
        Netcdf version to be used when creating the output
    window:
        On indices relying on a rolling window of days, configure the window width.
    threshold:
        On indices relying on a threshold, configure the threshold value. Unit less.
        The unit "degC" is added by icclim.
    transfer_limit_Mbytes:
        The dask maximum chunk size.
    out_unit:
        The output unit overriding Xclim results.
    callback:
        A callable to produce a progress bar
    """

    freq: Frequency
    _cf_variables: list[CfVariable]
    save_percentile: bool = False
    is_percent: bool = False
    netcdf_version: NetcdfVersion
    window: int | None
    threshold: float | None
    transfer_limit_Mbytes: int | None
    out_unit: str | None
    callback: Callable[[int], None] | None

    def __init__(
        self,
        ds: Dataset,
        slice_mode: SliceMode,
        var_names: list[str],
        netcdf_version: str | NetcdfVersion,
        index: Any | None,  # EcadIndex proper typing causes circular dependency
        save_percentile: bool = False,
        only_leap_years: bool = False,
        ignore_Feb29th: bool = False,
        window_width: int | None = 5,
        time_range: list[datetime] | None = None,
        base_period_time_range: list[datetime] | None = None,
        threshold: float | None = None,
        out_unit: str | None = None,
        interpolation: QuantileInterpolation
        | None = QuantileInterpolation.MEDIAN_UNBIASED,
        callback: Callable[[int], None] | None = None,
        chunk_it: bool = False,
    ):
        self.freq = Frequency.lookup(slice_mode)
        if time_range is not None:
            time_range = [x.strftime("%Y-%m-%d") for x in time_range]
        if base_period_time_range is not None:
            base_period_time_range = [
                x.strftime("%Y-%m-%d") for x in base_period_time_range
            ]
        self._cf_variables = [
            _build_cf_variable(
                da=ds[var_name],
                name=var_name,
                time_range=time_range,
                ignore_Feb29th=ignore_Feb29th,
                base_period_time_range=base_period_time_range,
                only_leap_years=only_leap_years,
                chunk_it=chunk_it,
                pre_processing=self.freq.pre_processing,
            )
            for var_name in var_names
        ]
        self.window = window_width
        self.save_percentile = save_percentile
        self.is_percent = out_unit == "%"
        self.out_unit = out_unit
        if isinstance(netcdf_version, str):
            self.netcdf_version = NetcdfVersion.lookup(netcdf_version)
        else:
            self.netcdf_version = netcdf_version
        self.interpolation = interpolation
        self.threshold = threshold
        self.callback = callback
        self.index = index

    @property
    def tas(self) -> CfVariable:
        tas_vars = list(filter(lambda v: v.name in TAS, self._cf_variables))
        if len(tas_vars) == 1:
            return tas_vars[0]
        # Otherwise rely on positional guess
        return self._cf_variables[0]

    @property
    def tasmax(self) -> CfVariable:
        tas_max_vars = list(filter(lambda v: v.name in TAS_MAX, self._cf_variables))
        if len(tas_max_vars) == 1:
            return tas_max_vars[0]
        # Otherwise rely on positional guess
        return self._cf_variables[0]

    @property
    def tasmin(self) -> CfVariable:
        tas_min_vars = list(filter(lambda v: v.name in TAS_MIN, self._cf_variables))
        if len(tas_min_vars) == 1:
            return tas_min_vars[0]
        # Otherwise rely on positional guess
        if len(self._cf_variables) > 1:
            # compound indices case
            return self._cf_variables[1]
        return self._cf_variables[0]

    @property
    def pr(self) -> CfVariable:
        pr_vars = list(filter(lambda v: v.name in PR, self._cf_variables))
        if len(pr_vars) == 1:
            return pr_vars[0]
        # Otherwise rely on positional guess
        if len(self._cf_variables) > 1:
            # compound indices case
            return self._cf_variables[1]
        return self._cf_variables[0]


def _build_cf_variable(
    da: DataArray,
    name: str,
    time_range: list[str] | None,
    ignore_Feb29th: bool,
    base_period_time_range: list[str] | None,
    only_leap_years: bool,
    chunk_it: bool,
    pre_processing: Callable,
) -> CfVariable:
    if chunk_it:
        da = da.chunk("auto")  # noqa - typing fixed in futur xarray version
    study_da = _build_study_da(da, time_range, ignore_Feb29th)
    study_da = pre_processing(study_da)
    if base_period_time_range is not None:
        reference_da = _build_reference_da(da, base_period_time_range, only_leap_years)
        reference_da = pre_processing(reference_da)
    else:
        reference_da = study_da
    # TODO: all these operations should probably be added in history metadata
    #       it could be a property in CfVariable which will be reused when we update the
    #       metadata of the index, at the end.
    return CfVariable(name, study_da, reference_da)


def _build_study_da(
    original_da: DataArray, time_range: list[str] | None, ignore_Feb29th: bool
) -> DataArray:
    if time_range is not None:
        if len(time_range) != 2:
            raise InvalidIcclimArgumentError(
                f"The given `time_range` {time_range}"
                f" has {len(time_range)} elements."
                f" It must have exactly 2 dates."
            )
        da = original_da.sel(time=slice(time_range[0], time_range[1]))
        if len(da.time) == 0:
            raise InvalidIcclimArgumentError(
                f"The given `time_range` {time_range} "
                f"is out of the dataset time period: "
                f"{original_da.time.min().dt.floor('D').values} "
                f"- {original_da.time.max().dt.floor('D').values}."
            )
    else:
        da = original_da
    if ignore_Feb29th:
        da = calendar.convert_calendar(da, CfCalendar.NO_LEAP.get_name())  # type:ignore
    return da


def _build_reference_da(
    original_da: DataArray,
    base_period_time_range: list[str],
    only_leap_years: bool,
) -> DataArray:
    # TODO merge with _build_data_array ?
    if len(base_period_time_range) != 2:
        raise InvalidIcclimArgumentError(
            f"The given `base_period_time_range` {base_period_time_range}"
            f" has {len(base_period_time_range)} elements."
            f" It must have exactly 2 dates."
        )
    da = original_da.sel(
        time=slice(base_period_time_range[0], base_period_time_range[1])
    )
    if len(da.time) == 0:
        raise InvalidIcclimArgumentError(
            f"The given `base_period_time_range` {base_period_time_range}"
            f" is out of the sample time bounds: "
            f"{original_da.time.min().dt.floor('D').values} "
            f"- {original_da.time.max().dt.floor('D').values}."
        )
    if only_leap_years:
        da = _reduce_only_leap_years(original_da)
    return da


def _reduce_only_leap_years(da: DataArray) -> DataArray:
    reduced_list: list[DataArray] = []
    for _, val in da.groupby(da.time.dt.year):
        if val.time.dt.dayofyear.max() == 366:
            reduced_list.append(val)
    if not reduced_list:
        raise InvalidIcclimArgumentError(
            "No leap year in current dataset. Do not use only_leap_years parameter."
        )
    return xarray.concat(reduced_list, "time")



# keep imports below to expose api in `icclim` namespace
from ._generated_api import *  # noqa
from .main import index, indice, indices  # noqa
from .pre_processing.rechunk import create_optimized_zarr_store  # noqa

__version__ = "5.2.1"



from __future__ import annotations

from enum import Enum
from typing import Any


class IndexGroup(Enum):
    TEMPERATURE = "temperature"
    HEAT = "heat"
    COLD = "cold"
    DROUGHT = "drought"
    RAIN = "rain"
    SNOW = "snow"
    COMPOUND = "compound"
    WILD_CARD_GROUP = "all"  # no index bound to it

    @staticmethod
    def lookup(query: str | IndexGroup) -> IndexGroup:
        if isinstance(query, IndexGroup):
            return query
        for gr in IndexGroup:
            if gr.value.upper() == query.upper():
                return gr
        valid_values = list(map(lambda x: x.value, IndexGroup))
        raise NotImplementedError(
            f"IndexGroup must be one of the following: {valid_values},"
            f" but query was {query}."
        )

    def get_indices(self) -> list[Any]:
        # import locally to avoid circular dependency (an index has already a group)
        from icclim.models.ecad_indices import EcadIndex

        return list(filter(lambda i: i.group == self, EcadIndex))



from __future__ import annotations

import datetime
from typing import Literal, TypedDict

from icclim.models.user_index_config import LogicalOperationLiteral
from icclim.user_indices.calc_operation import CalcOperation, CalcOperationLiteral


class UserIndexDict(TypedDict, total=False):
    index_name: str
    calc_operation: CalcOperationLiteral | CalcOperation
    logical_operation: LogicalOperationLiteral | None
    thresh: str | float | None
    link_logical_operations: Literal["and", "or"] | None
    extreme_mode: Literal["min", "max"] | None
    window_width: int | None
    coef: float | None
    date_event: bool | None
    var_type: Literal["t", "p"] | None
    ref_time_range: list[datetime.datetime] | None  # length of 2
    # deprecated
    indice_name: str | None



from unittest.mock import MagicMock, patch

import numpy as np
import pandas as pd
import pytest
import xarray as xr

from icclim import create_optimized_zarr_store
from icclim.icclim_exceptions import InvalidIcclimArgumentError


def test_create_optimized_zarr_store_success():
    ds = xr.Dataset(
        {
            "tas": xr.DataArray(
                data=np.full(10, 42).reshape((10, 1, 1)),
                coords=dict(
                    lat=[42],
                    lon=[42],
                    time=pd.date_range("2042-01-01", periods=10, freq="D"),
                ),
                dims=["time", "lat", "lon"],
                name="pr",
                attrs={"units": "kg m-2 d-1"},
            )
        }
    ).chunk({"time": 2})
    with create_optimized_zarr_store(
        in_files=ds,
        var_names="tas",
        target_zarr_store_name="yolo.zarr",
    ) as result:
        assert len(result.chunks["time"]) == 1
        np.testing.assert_array_equal(result.tas.values, ds.tas.values)
        np.testing.assert_array_equal(result.data_vars.keys(), ds.data_vars.keys())


def test_create_optimized_zarr_store_error():
    ds = xr.Dataset(
        {
            "tas": xr.DataArray(
                data=np.full(10, 42).reshape((10, 1, 1)),
                coords=dict(
                    lat=[42],
                    lon=[42],
                    time=pd.date_range("2042-01-01", periods=10, freq="D"),
                ),
                dims=["time", "lat", "lon"],
                name="pr",
                attrs={"units": "kg m-2 d-1"},
            )
        }
    ).chunk({"time": 2})
    # Then
    with pytest.raises(InvalidIcclimArgumentError):
        # When
        with create_optimized_zarr_store(
            in_files=ds,
            var_names="TATAYOYO!",
            target_zarr_store_name="yolo.zarr",
        ):
            pass


@patch("rechunker.rechunk")
def test_create_optimized_zarr_store_no_rechunk(rechunk_mock: MagicMock):
    ds = xr.Dataset(
        {
            "tas": xr.DataArray(
                data=np.full(10, 42).reshape((10, 1, 1)),
                coords=dict(
                    lat=[42],
                    lon=[42],
                    time=pd.date_range("2042-01-01", periods=10, freq="D"),
                ),
                dims=["time", "lat", "lon"],
                name="pr",
                attrs={"units": "kg m-2 d-1"},
            )
        }
    ).chunk({"time": 2})
    # When
    with pytest.raises(InvalidIcclimArgumentError):
        with create_optimized_zarr_store(
            in_files=ds,
            var_names="tas",
            target_zarr_store_name="n/a",
            chunking={"time": 2},
        ):
            pass
    rechunk_mock.assert_not_called()



from datetime import datetime
from unittest.mock import MagicMock, patch

import numpy as np
import pytest

import icclim
from icclim.icclim_logger import Verbosity
from icclim.models.constants import (
    MODIFIABLE_QUANTILE_WINDOW,
    MODIFIABLE_THRESHOLD,
    MODIFIABLE_UNIT,
    QUANTILE_BASED,
)
from icclim.models.ecad_indices import EcadIndex
from icclim.models.frequency import Frequency
from icclim.models.netcdf_version import NetcdfVersion
from icclim.models.quantile_interpolation import QuantileInterpolation
from icclim.tests.test_utils import stub_tas
from icclim.user_indices.calc_operation import CalcOperation

DEFAULT_ARGS = dict(
    in_files="pouet.nc",
    var_name=None,
    slice_mode=Frequency.YEAR,
    time_range=None,
    out_file=None,
    ignore_Feb29th=False,
    netcdf_version=NetcdfVersion.NETCDF4,
    logs_verbosity=Verbosity.LOW,
)


def build_expected_args(index):
    expected_call_args = {"index_name": index.name}
    expected_call_args.update(DEFAULT_ARGS)
    if MODIFIABLE_THRESHOLD in index.qualifiers:
        expected_call_args.update({"threshold": None})
    if QUANTILE_BASED in index.qualifiers:
        expected_call_args.update(
            {
                "base_period_time_range": None,
                "only_leap_years": False,
                "interpolation": QuantileInterpolation.MEDIAN_UNBIASED,
                "save_percentile": False,
            }
        )
    if MODIFIABLE_QUANTILE_WINDOW in index.qualifiers:
        expected_call_args.update({"window_width": 5})
    if MODIFIABLE_UNIT in index.qualifiers:
        expected_call_args.update({"out_unit": None})

    return expected_call_args


@patch("icclim.index")
def test_generated_api(generic_index_fun_mock: MagicMock):
    for i in EcadIndex:
        print(i)
        # GIVEN
        api_index_fun = eval(f"icclim.{i.name.lower()}")
        # WHEN
        api_index_fun(**DEFAULT_ARGS)
        # THEN
        expected_call_args = build_expected_args(i)
        generic_index_fun_mock.assert_called_with(**expected_call_args)


@patch("icclim.index")
def test_custom_index(index_fun_mock: MagicMock):
    user_index_args = dict(
        in_files="pouet_file.nc",
        var_name=None,
        slice_mode=Frequency.YEAR,
        time_range=None,
        out_file=None,
        base_period_time_range=None,
        only_leap_years=False,
        ignore_Feb29th=False,
        out_unit=None,
        netcdf_version=NetcdfVersion.NETCDF4,
        save_percentile=False,
        logs_verbosity=Verbosity.LOW,
        user_index={
            "index_name": "pouet",
            "calc_operation": "nb_events",
            "logical_operation": "gt",
            "thresh": 0,
            "date_event": True,
        },
    )
    icclim.custom_index(**user_index_args)
    index_fun_mock.assert_called_with(**user_index_args)


# integration test
def test_txx__season_slice_mode():
    tas = stub_tas()
    tas.loc[{"time": "2042-02-02"}] = 295
    tas.loc[{"time": "2042-01-01"}] = 303.15  # 30ºC 273.15
    res = icclim.txx(tas, slice_mode=["season", [11, 12, 1, 2]]).compute()
    np.testing.assert_array_equal(res.TXx.isel(time=0), 30)
    np.testing.assert_array_equal(
        res.time_bounds.isel(time=0),
        [np.datetime64("2041-11-01"), np.datetime64("2042-02-28")],
    )


def test_txx__months_slice_mode():
    tas = stub_tas()
    tas.loc[{"time": "2042-11-02"}] = 295
    tas.loc[{"time": "2042-01-01"}] = 303.15  # 30ºC 273.15
    res = icclim.txx(tas, slice_mode=["months", [11, 1]]).compute()
    np.testing.assert_array_equal(res.TXx.isel(time=0), 30)
    np.testing.assert_almost_equal(res.TXx.isel(time=1), 21.85)
    np.testing.assert_array_equal(
        res.time_bounds.isel(time=0),
        [np.datetime64("2042-01-01"), np.datetime64("2042-01-31")],
    )


# integration test
@pytest.mark.parametrize(
    "operator, exp_y1, exp_y2",
    [
        (CalcOperation.MIN, 303.15, 280.15),
        (CalcOperation.MAX, 303.15, 280.15),
        (CalcOperation.SUM, 303.15, 280.15),  # values below 275 are filtered out
        (CalcOperation.MEAN, 303.15, 280.15),
        (CalcOperation.EVENT_COUNT, 1, 1),
        (CalcOperation.MAX_NUMBER_OF_CONSECUTIVE_EVENTS, 1, 1),
    ],
)
def test_custom_index__season_slice_mode(operator, exp_y1, exp_y2):
    tas = stub_tas(2.0)
    tas.loc[{"time": "2042-01-01"}] = 303.15
    tas.loc[{"time": "2042-12-01"}] = 280.15
    res = icclim.custom_index(
        in_files=tas,
        slice_mode=["season", [12, 1]],
        var_name="a_name",
        user_index={
            "index_name": "pouet",
            "calc_operation": operator,
            "logical_operation": "gt",
            "thresh": 275,
        },
    ).compute()
    np.testing.assert_almost_equal(res.pouet.isel(time=0), exp_y1)
    np.testing.assert_almost_equal(res.pouet.isel(time=1), exp_y2)


# integration test
@pytest.mark.parametrize(
    "operator, exp_y1, exp_y2",
    [
        (CalcOperation.RUN_MEAN, 2, 2),
        (CalcOperation.RUN_SUM, 14, 14),
    ],
)
def test_custom_index_run_algos__season_slice_mode(operator, exp_y1, exp_y2):
    tas = stub_tas(2.0)
    res = icclim.custom_index(
        in_files=tas,
        slice_mode=["season", [12, 1]],
        var_name="a_name",
        user_index={
            "index_name": "pouet",
            "calc_operation": operator,
            "extreme_mode": "max",
            "window_width": 7,
        },
    ).compute()
    np.testing.assert_almost_equal(res.pouet.isel(time=0), exp_y1)
    np.testing.assert_almost_equal(res.pouet.isel(time=1), exp_y2)


def test_custom_index_anomaly__season_slice_mode():
    tas = stub_tas(2.0)
    tas.loc[{"time": "2045-01-01"}] = 300
    res = icclim.custom_index(
        in_files=tas,
        slice_mode=["season", [12, 1]],
        var_name="a_name",
        user_index={
            "index_name": "anomaly",
            "calc_operation": CalcOperation.ANOMALY,
            "ref_time_range": [datetime(2042, 1, 1), datetime(2044, 12, 31)],
        },
    ).compute()
    np.testing.assert_almost_equal(res.anomaly, 0.96129032)



# fmt: off
# flake8: noqa

# placeholder for user_indices
PERCENTILE_THRESHOLD_STAMP = "p"
WET_DAY_THRESHOLD = 1  # 1mm
PRECIPITATION = "p"
TEMPERATURE = "t"

# percentiles dimension from percentile_doy
PERCENTILES_COORD = "percentiles"
# attribut holding the in_base time bounds
IN_BASE_IDENTIFIER = "reference_epoch"

# Aliases of input variables names. Source: clix-meta
PR = ["pr", "pradjust", "prec", "rr", "precip", "PREC", "Prec", "RR", "PRECIP", "Precip"]
TAS = ["tas", "tavg", "ta", "tasadjust", "tmean", "tm", "tg", "meant", "TMEAN", "Tmean", "TM", "TG", "MEANT", "meanT", "tasmidpoint"]
TAS_MAX = ["tasmax", "tasmaxadjust", "tmax", "tx", "maxt", "TMAX", "Tmax", "TX", "MAXT", "maxT"]
TAS_MIN = ["tasmin", "tasminadjust", "tmin", "tn", "mint", "TMIN", "Tmin", "TN", "MINT", "minT"]

# Source of index definition
ECAD_ATBD = "ECA&D, Algorithm Theoretical Basis Document (ATBD) v11"

# Index qualifiers
QUANTILE_BASED = "QUANTILE_BASED"  # fields: QUANTILE_INDEX_FIELDS
MODIFIABLE_UNIT = "MODIFIABLE_UNIT"  # fields: out_unit
MODIFIABLE_THRESHOLD = "MODIFIABLE_THRESHOLD"  # fields: threshold
MODIFIABLE_QUANTILE_WINDOW = "MODIFIABLE_QUANTILE_WINDOW"  # fields: window_width

# Map of months index to their short name, used to get a pandas frequency anchor
MONTHS_MAP = {1:"JAN",  2:"FEB", 3:"MAR", 4:"APR", 5:"MAY", 6:"JUN", 7:"JUL", 8:"AUG", 9:"SEP", 10:"OCT", 11:"NOV", 12:"DEC" }

# Season defined by their month numbers
AMJJAS_MONTHS = [*range(4, 9)]
ONDJFM_MONTHS = [10, 11, 12, 1, 2, 3]
DJF_MONTHS = [12, 1, 2]
MAM_MONTHS = [*range(3, 6)]
JJA_MONTHS = [*range(6, 9)]
SON_MONTHS = [*range(9, 12)]



import os
from unittest.mock import MagicMock, patch

import cftime
import numpy as np
import pandas as pd
import pytest
import xarray as xr

import icclim
from icclim.models.ecad_indices import EcadIndex
from icclim.models.frequency import Frequency
from icclim.models.index_group import IndexGroup


@patch("icclim.main.index")
@patch("icclim.icclim_logger.IcclimLogger")
def test_deprecated_indice(log_mock: MagicMock, index_mock: MagicMock):
    icclim.main.log = log_mock
    icclim.indice()
    log_mock.deprecation_warning.assert_called_once_with(
        old="icclim.indice", new="icclim.index"
    )
    index_mock.assert_called_once()


HEAT_INDICES = ["SU", "TR", "WSDI", "TG90p", "TN90p", "TX90p", "TXx", "TNx", "CSU"]


@pytest.mark.slow
class Test_Integration:
    """
    Simple integration test.
    We are not testing here the actual indices results, they are already tested in
    `test_ecad_indices.py` as well as in xclim directly.
    The goal it to make sure every the whole app can run smoothly

    These tests have side effect:
    - writing and removing of "out.nc" file

    """

    OUTPUT_FILE = "out.nc"
    CF_TIME_RANGE = pd.date_range(start="2042-01-01", end="2045-12-31", freq="D")
    TIME_RANGE = xr.cftime_range("2042-01-01", end="2045-12-31", freq="D")
    data = xr.DataArray(
        data=(np.full(len(TIME_RANGE), 20).reshape((len(TIME_RANGE), 1, 1))),
        dims=["time", "lat", "lon"],
        coords=dict(
            lat=[42],
            lon=[42],
            time=TIME_RANGE,
        ),
        attrs={"units": "degC"},
    )

    data_cf_time = xr.DataArray(
        data=(np.full(len(TIME_RANGE), 20).reshape((len(TIME_RANGE), 1, 1))),
        dims=["time", "lat", "lon"],
        coords=dict(
            lat=[42],
            lon=[42],
            time=CF_TIME_RANGE,
        ),
        attrs={"units": "degC"},
    )

    @pytest.fixture(autouse=True)
    def cleanup(self):
        # setup
        # ...
        yield
        # teardown
        try:
            os.remove(self.OUTPUT_FILE)
        except FileNotFoundError:
            pass

    def test_index_SU(self):
        res = icclim.index(
            indice_name="SU", in_files=self.data, out_file=self.OUTPUT_FILE
        )
        np.testing.assert_array_equal(0, res.SU)

    def test_index_SU__monthy_sampled(self):
        res = icclim.index(
            indice_name="SU",
            in_files=self.data,
            out_file=self.OUTPUT_FILE,
            slice_mode=Frequency.MONTH,
        )
        np.testing.assert_array_equal(0, res.SU)
        np.testing.assert_array_equal(
            len(np.unique(self.TIME_RANGE.year)) * 12, len(res.time)
        )

    def test_index_SU__monthy_sampled_cf_time(self):
        res = icclim.index(
            indice_name="SU",
            in_files=self.data,
            out_file=self.OUTPUT_FILE,
            slice_mode=Frequency.MONTH,
        )
        np.testing.assert_array_equal(0, res.SU)
        res.time_bounds.isel(time=0)
        np.testing.assert_array_equal(
            len(np.unique(self.TIME_RANGE.year)) * 12, len(res.time)
        )
        assert res.time_bounds.sel(time=res.time[0])[0] == cftime.DatetimeGregorian(
            2042, 1, 1, 0, 0, 0, 0
        )
        assert res.time_bounds.sel(time=res.time[0])[1] == cftime.DatetimeGregorian(
            2042, 1, 31, 0, 0, 0, 0
        )

    def test_index_SU__DJF_cf_time(self):
        res = icclim.index(
            indice_name="SU",
            in_files=self.data,
            out_file=self.OUTPUT_FILE,
            slice_mode=Frequency.DJF,
        )
        np.testing.assert_array_equal(0, res.SU)
        # 1 more year as DJF sampling create a months withs nans before
        np.testing.assert_array_equal(
            len(np.unique(self.TIME_RANGE.year)) + 1, len(res.time)
        )
        assert res.time_bounds.sel(time=res.time[0])[0] == cftime.DatetimeGregorian(
            2041, 12, 1, 0, 0, 0, 0
        )
        assert res.time_bounds.sel(time=res.time[0])[1] == cftime.DatetimeGregorian(
            2042, 2, 28, 0, 0, 0, 0
        )

    def test_indices_from_DataArray(self):
        res = icclim.indices(
            index_group=IndexGroup.HEAT, in_files=self.data, out_file=self.OUTPUT_FILE
        )
        for i in HEAT_INDICES:
            assert res[i] is not None

    def test_indices_all_from_Dataset(self):
        ds = self.data.to_dataset(name="tas")
        ds["tasmax"] = self.data
        ds["tasmin"] = self.data
        ds["pr"] = self.data.copy(deep=True)
        ds["pr"].attrs["units"] = "kg m-2 d-1"
        ds["prec"] = self.data.copy(deep=True)
        ds["prec"].attrs["units"] = "cm"
        res = icclim.indices(
            index_group="all", in_files=ds, out_file=self.OUTPUT_FILE, ignore_error=True
        )
        for i in EcadIndex:
            assert res[i.short_name] is not None

    def test_indices_all_ignore_error(self):
        ds = self.data.to_dataset(name="tas")
        ds["tasmax"] = self.data
        ds["tasmin"] = self.data
        ds["pr"] = self.data.copy(deep=True)
        ds["pr"].attrs["units"] = "kg m-2 d-1"
        res: xr.Dataset = icclim.indices(
            index_group="all", in_files=ds, out_file=self.OUTPUT_FILE, ignore_error=True
        )
        for i in EcadIndex:
            # No variable in input to compute for snow indices
            if i.group == IndexGroup.SNOW:
                assert res.data_vars.get(i.short_name, None) is None
            else:
                assert res[i.short_name] is not None

    def test_indices_all_error(self):
        ds = self.data.to_dataset(name="tas")
        ds["tasmax"] = self.data
        ds["tasmin"] = self.data
        ds["pr"] = self.data.copy(deep=True)
        ds["pr"].attrs["units"] = "kg m-2 d-1"
        with pytest.raises(Exception):
            icclim.indices(
                index_group="all",
                in_files=ds,
                out_file=self.OUTPUT_FILE,
                ignore_error=False,
            )
